@@ -35,6 +35,7 @@ public:
 public Q_SLOTS:
     QDBusPendingReply<> focusIn();
     QDBusPendingReply<> focusOut();
+    QDBusPendingReply<> hideVirtualKeyboard();
     QDBusPendingReply<bool> processKeyEvent(unsigned int keyval,
                                             unsigned int keycode,
                                             unsigned int state, bool type,
@@ -50,10 +51,12 @@ public Q_SLOTS:
                                            unsigned int anchor);
     QDBusPendingReply<> setSurroundingTextPosition(unsigned int cursor,
                                                    unsigned int anchor);
+    QDBusPendingReply<> showVirtualKeyboard();
     QDBusPendingReply<> prevPage();
     QDBusPendingReply<> nextPage();
     QDBusPendingReply<> selectCandidate(int i);
     QDBusPendingReply<> invokeAction(unsigned int action, int cursor);
+    bool isVirtualKeyboardVisible();
 
     bool supportInvokeAction() const;
 
@@ -73,6 +76,9 @@ Q_SIGNALS:
                             int candidateIndex, int layoutHint, bool hasPrev,
                             bool hasNext);
     void inputContextCreated(const QByteArray &uuid);
+    void inputContextCreationFailed();
+    void notifyFocusOut();
+    void virtualKeyboardVisibilityChanged(bool visible);
 
 private:
     FcitxQtInputContextProxyPrivate *const d_ptr;
